from typing import Optional, List

from . import exceptions
from ._translators import TranslatedTimestamp, windows_bytes_to_unix_date
from ._attribute_codecs import ObjectGUIDCodec, ObjectSidCodec, ThumbnailPhotoCodec, LogonHoursCodec


def string(value):
    def _cast(value):
        try:
            return {bytes: lambda x: x.decode("utf-8"), int: lambda x: str(x), str: lambda x: x}[type(value)](value)
        except KeyError:
            raise exceptions.AttributeDecodingError(
                f"Cannot create a string attribute from {value} of type {type(value)}"
            )

    if all([isinstance(value, list), len(value) > 1]):
        raise exceptions.AttributeDecodingError(
            f"Cannot create a string attribute value from a list with multiple items: {value}"
        )

    try:
        return _cast(value[0])
    except KeyError:
        raise exceptions.NoSuchObject("Failed to create a string attribute. No values in LDAP attribute list.")


def optional_string(value):
    try:
        return string(value)
    except exceptions.NoSuchObject:
        # None is valid in an Optional[str]
        return None


def string_list(value):
    def _cast(value):
        try:
            return {bytes: lambda x: x.decode("utf-8"), int: lambda x: str(x), str: lambda x: x}[type(value)](value)
        except KeyError:
            raise exceptions.AttributeDecodingError(f"Cannot decode list item {value} of type {type(value)} to string.")

    return [_cast(v) for v in value]


def optional_string_list(value):
    return string_list(value) if value else None


def integer(value):

    if all([isinstance(value, list), len(value) > 1]):
        raise exceptions.AttributeDecodingError(
            f"Cannot create an integer attribute value from a list with multiple items: {value}"
        )

    def _cast(value):
        try:
            return {bytes: lambda x: int(x.decode("utf-8")), int: lambda x: x, str: lambda x: int(x)}[type(value)](
                value
            )
        except KeyError:
            raise exceptions.AttributeDecodingError(
                f"Cannot create a single integer attribute from {value} of type {type(value)}"
            )

    try:
        return _cast(value[0])
    except KeyError:
        raise exceptions.NoSuchObject("Failed to create an integer attribute. No values in LDAP attribute list.")


def optional_integer(value):
    try:
        return integer(value)
    except KeyError:
        # None is valid in an Optional[int]
        return None


def timestamp(value):
    if len(value) > 1:
        raise exceptions.AttributeDecodingError(
            f"Cannot create an integer attribute value from a list with multiple items: {value}"
        )
    try:
        return windows_bytes_to_unix_date(value[0])
    except KeyError:
        raise exceptions.NoSuchObject("Failed to create a timestamp attribute. No values in LDAP attribute list.")


def optional_timestamp(value):
    try:
        return windows_bytes_to_unix_date(value[0])
    except KeyError:
        # None is valid in an Optional[TranslatedTimestamp]
        return None


def object_sid(value):
    try:
        return ObjectSidCodec.decode(value[0])
    except IndexError:
        raise exceptions.AttributeDecodingError("LDAP object is missing mandatory attribute 'objectSid'.")


def object_guid(value):
    try:
        return ObjectGUIDCodec.decode(value[0])
    except IndexError:
        raise exceptions.AttributeDecodingError("LDAP object is missing mandatory attribute 'objectGUID'.")


def logon_hours(value):
    return LogonHoursCodec.decode(value[0]) if value[0] else None


def thumbnail_photo(value):
    return ThumbnailPhotoCodec.decode(value[0]) if value[0] else None


generic_field_functions = {
    "objectSid": object_sid,
    "objectGUID": object_guid,
    "logonHours": logon_hours,
    "thumbnailPhoto": thumbnail_photo,
}

generic_type_functions = {
    str: string,
    Optional[str]: optional_string,
    int: integer,
    Optional[int]: optional_integer,
    List[str]: string_list,
    Optional[List[str]]: optional_string_list,
    Optional[TranslatedTimestamp]: optional_timestamp,
}
