"""
Organizational Units

This module contains actions related to Active Directory Organizational Units.
"""
import uuid
import logging

from typing import List, Dict, Optional, Union, Any


from ppa import ui
from active_directory import organizational_units, exceptions as sdk_exceptions
from playbook_tools import tables, process

from .exceptions import NoOrganizationalUnitFound, InvalidAttribute
from ._wrappers import ldap_function, idempotent_create, idempotent_remove, process_organizational_unit, bool_to_search_scope
from .records import DomainController, OrganizationalUnit


OU_TABLE_HEADER = ["Name", "Distinguished Name"]
OU_TABLE_FIELDS = ["name", "distinguishedName"]


logger = logging.getLogger(__name__)


@ldap_function
def get_all(
    domain_controller: DomainController, search_base: Optional[str] = None, one_level: Optional[bool] = False
) -> List[OrganizationalUnit]:
    """Get all organizational units.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 6.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A list of [Organizational Units](../inputs_outputs/#organizationalunit).

??? example
    ```YAML linenums="1"
    - active_directory.organizational_units.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_ous
    ```
"""
    logger.info(f"getting all organizational units using search base {search_base if search_base else 'default'}")
    return list(
        map(
            process_organizational_unit,
            organizational_units.get_all(
                domain_controller, search_base=search_base, search_scope=bool_to_search_scope(one_level)
            ),
        )
    )


@ldap_function
def by_distinguishedname(domain_controller: DomainController, distinguishedName: str) -> OrganizationalUnit:
    """Get an organizational unit by its 'distinguishedName' attribute.

!!! question "Minimum Plugin Version: 5.2.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ The organizational unit's `distinguishedName`

??? success "Output"

    A single [Organizational Unit](../inputs_outputs/#organizationalunit).

??? example
    ```YAML linenums="1"
    - active_directory.organizational_units.by_distinguishedname:
        distinguishedName: OU=Engineers,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller_info
      save: example_ou
    ```
"""
    logger.info(f"getting organizational unit with distinguishedName '{distinguishedName}'")
    try:
        return process_organizational_unit(organizational_units.search(domain_controller, distinguishedName=distinguishedName)[0])
    except IndexError:
        raise NoOrganizationalUnitFound(f"No organizational unit found with distinguishedName '{distinguishedName}'.")


@ldap_function
def search(
    domain_controller: DomainController,
    search_params: Dict[str, Any],
    search_base: Optional[str] = None,
    one_level: Optional[bool] = False,
) -> process.SearchResult:
    """Search for organizational units using LDAP attributes & values.

!!! question "Minimum Plugin Version: 6.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller)

    - __search_params:__ a dictionary containing [Organizational Unit](../inputs_outputs#organizationalunit) keys & values to use in the search

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Organizational Units](../inputs_outputs/#organizationalunit) found

    - **all**: a list of [Organizational Units](../inputs_outputs/#organizationalunit) found

    - **first**: the first [Organizational Unit](../inputs_outputs/#organizationalunit) found

    - **last**: the last [Organizational Unit](../inputs_outputs/#organizationalunit) found

??? warning "Search Result Format"

    The search result format was updated in version 6.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Searching for all organizational units in the __Users__ CN whose `cn` starts with __admin*__:

    ```YAML linenums="1"
    - active_directory.organizational_units.search:
        search_params:
          cn: admin*
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: organizational_units
    ```

!!! tip

    You can use `*` as a wildcard at the end of search values.
"""
    logger.info(
        f"searching for organizational units in search base {search_base if search_base else 'default'} using the following parameters: {search_params}"
    )
    return process.search_result(
        list(
            map(
                process_organizational_unit,
                organizational_units.search(
                    domain_controller, search_base=search_base, search_scope=bool_to_search_scope(one_level), **search_params
                ),
            )
        )
    )


@ldap_function
@idempotent_create
def create(
    domain_controller: DomainController, name: str, distinguishedName: str, extra_params: Optional[Dict[str, Any]] = None,
) -> None:
    """Create a new organizational unit.

!!! question "Minimum Plugin Version: 5.2.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller)

    - __name:__ the new organizational unit name

    - __distinguishedName:__ the new organizational unit distinguishedName

    - __extra_params:__ a dictionary containing any extra LDAP attributes & values for the organizational unit

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.organizational_units.create:
        name: example.ou
        distinguishedName: "OU=Example OU,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """

    logger.info(
        f"creating organizational unit '{name}' at '{distinguishedName}' with the following extra parameters: {extra_params}"
    )
    try:
        return organizational_units.create(
            domain_controller, distinguished_name=distinguishedName, name=name, **extra_params or {},
        )
    except sdk_exceptions.ParameterError as e:
        logger.error(e)
        message = "One or more of the following is not a supported organizational unit attribute: {}.".format(
            ", ".join([f"'{key}'" for key in list(extra_params.keys())])
        )
        ui.output_error(message)
        raise InvalidAttribute(message)


@ldap_function
@idempotent_remove
def delete(domain_controller: DomainController, distinguishedName: str) -> None:
    """Delete an organizational unit.

!!! question "Minimum Plugin Version: 5.2.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller)

    - __distinguishedName:__ the distinguishedName of the organizational unit to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.organizational_units.delete:
        distinguishedName: "OU=Engineers,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    logger.info(f"deleting organizational unit with distinguished name {distinguishedName}")
    return organizational_units.delete(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def move(domain_controller: DomainController, distinguishedName: str, parent_distinguishedName: str) -> None:
    """Move an organizational unit to a different OU or CN.

!!! question "Minimum Plugin Version: 5.2.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller)

    - __distinguishedName:__ the distinguishedName of the organizational unit to move

    - __parent_distinguishedName:__ the distinguishedName of the new organizational unit parent

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.organizational_units.move:
        distinguishedName: "OU=Engineers,CN=Users,DC=Example,DC=Domain"
        parent_distinguishedName: "OU=Staff,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return organizational_units.move(domain_controller, distinguished_name=distinguishedName, new_parent=parent_distinguishedName)


def output_table(text: str, organizational_units: Union[OrganizationalUnit, List[OrganizationalUnit]]) -> None:
    """Display organizational units in a table.

The table will have the following columns:

- Name
- Distinguished Name

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 5.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __organizational_units:__ a single or list of [Organizational Units](../inputs_outputs/#organizationalunit)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.organizational_units.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_ous

    - active_directory.organizational_units.output_table:
        text: All Organizational Units
      load:
        organizational_units: all_ous
    ```
    """
    organizational_units = [organizational_units] if not isinstance(organizational_units, list) else organizational_units
    ui.output_table(
        text,
        header=OU_TABLE_HEADER,
        data=[
            [str(ou.get(field)) if ou.get(field) is not None else "" for field in OU_TABLE_FIELDS] for ou in organizational_units
        ],
    )


def display(text: str, organizational_units: Union[OrganizationalUnit, List[OrganizationalUnit]]) -> None:
    """Display organizational units in a table.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __organizational_units:__ any number of [Organizational Units](../inputs_outputs/#organizationalunit)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.organizational_units.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_ous

    - active_directory.organizational_units.display:
        text: All Organizational Units
      load:
        organizational_units: all_ous
    ```
    """
    tables.output_table(text, header=OU_TABLE_HEADER, extract=OU_TABLE_FIELDS, records=organizational_units)


def display_tabbed(text: str, tabs: Dict[str, Union[OrganizationalUnit, List[OrganizationalUnit]]]) -> None:
    """Display organizational units in a table with multiple tabs.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [Organizational Units](../inputs_outputs/#organizationalunit)

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing user accounts from 2 OUs & presenting them in a tabbed table:

    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - active_directory.organizational_units.get_all:
      load:
        domain_controller: production_domain
      save: production_ous

    - active_directory.organizational_units.get_all:
      load:
        domain_controller: uat_domain
      save: uat_ous

    - ppa_tools.dictionaries.insert:
        name: Production
      load:
        value: production_ous
        dictionary: tabs
      save: tabs

    - ppa_tools.dictionaries.insert:
        name: UAT
      load:
        value: uat_ous
        dictionary: tabs
      save: tabs

    - active_directory.organizational_units.display_tabbed:
        text: Organizational Units
      load:
        tabs: tabs
    ```
    """
    tables.output_tabbed_table(text, header=OU_TABLE_HEADER, extract=OU_TABLE_FIELDS, tabs=tabs)


def output_tabbed_table(text: str, tabs: Dict[str, List[OrganizationalUnit]]) -> None:
    """Display an organizational units table with multiple tabs.

This action can be used to display organizational unit lists from multiple Active Directories.

The table will have the following columns:

- Name
- Distinguished Name

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display_tabbed](#display_tabbed).

!!! question "Minimum Plugin Version: 5.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is a list of [OrganizationalUnit](../inputs_outputs/#organizationalunit) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing user accounts from 2 domains & presenting them in a tabbed table:

    ```YAML linenums="1"
    - set:
        name: tabbed_table
        value: {}

    - active_directory.organizational_units.get_all:
      load:
        domain_controller: production_domain
      save: production_ous

    - active_directory.organizational_units.get_all:
      load:
        domain_controller: uat_domain
      save: uat_ous

    - ppa_tools.dictionaries.insert:
        name: Production Active Directory
      load:
        value: production_ous
        dictionary: tabbed_table
      save: tabbed_table

    - ppa_tools.dictionaries.insert:
        name: UAT Active Directory
      load:
        value: uat_ous
        dictionary: tabbed_table
      save: tabbed_table

    - active_directory.organizational_units.output_tabbed_table:
        text: Active Directory Organizational Units
      load:
        tabs: tabbed_table
    ```
    """
    ui.output_table(
        text,
        header=OU_TABLE_HEADER,
        data={
            key: [
                [str(ou.get(field)) if ou.get(field) is not None else "" for field in OU_TABLE_FIELDS]
                for ou in organizational_units
            ]
            for key, organizational_units in tabs.items()
        },
    )


def input_table(
    text: str,
    organizational_units: Union[OrganizationalUnit, List[OrganizationalUnit]],
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
) -> List[OrganizationalUnit]:
    """Display organizational units in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Distinguished Name

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 5.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __organizational_units:__ any number of [Organizational Units](../inputs_outputs/#organizationalunit)

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A list of [Organizational Units](../inputs_outputs/#organizationalunit).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example
    - Getting all organizational units whose common names start with _UAT_

    - Saving the results as a new variable called `uat_ous`

    - Using this action to show the `uat_ous` in a table, requiring the task operator selects at least 1

    - The selection is saved as a new variable called `selected_ous`

    ```YAML linenums="1"
    - active_directory.organizational_units.search:
        name: UAT
      load:
        domain_controller: domain_controller_info
      save: uat_ous

    - active_directory.organizational_units.input_table:
        text: "Please Select >= 1 Organizational Unit(s)"
        minimum: 1
      load:
        organizational_units: test_ous
      save: selected_ous
    ```
    """
    organizational_units = [organizational_units] if not isinstance(organizational_units, list) else organizational_units
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=OU_TABLE_HEADER,
            data=[
                [str(ou.get(field)) if ou.get(field) is not None else "" for field in OU_TABLE_FIELDS]
                for ou in organizational_units
            ],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [organizational_units[i] for i in indices]


def select(
    text: str,
    organizational_units: Union[OrganizationalUnit, List[OrganizationalUnit]],
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display organizational units in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __organizational_units:__ any number of [Organizational Units](../inputs_outputs/#organizationalunit)

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Organizational Units](../inputs_outputs/#organizationalunit)

    - **all**: a list of selected [Organizational Units](../inputs_outputs/#organizationalunit)

    - **first**: the first selected [Organizational Unit](../inputs_outputs/#organizationalunit)

??? example

    ```YAML linenums="1"
    - active_directory.organizational_units.get_all:
      load:
        domain_controller: domain_controller
      save: ous

    - active_directory.organizational_units.select:
        text: Select Organizational Units
      load:
        organizational_units: ous
      save: selection
    ```
    """
    return tables.input_table(
        text, header=OU_TABLE_HEADER, extract=OU_TABLE_FIELDS, records=organizational_units, minimum=minimum, maximum=maximum
    )


def select_one(text: str, organizational_units: Union[OrganizationalUnit, List[OrganizationalUnit]]) -> OrganizationalUnit:
    """Display Organizational Units in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __organizational_units:__ any number of [Organizational Units](../inputs_outputs/#organizationalunit)

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A single [Organizational Unit](../inputs_outputs/#organizationalunit).

??? example

    ```YAML linenums="1"
    - active_directory.organizational_units.get_all:
      load:
        domain_controller: domain_controller
      save: ous

    - active_directory.organizational_units.select_one:
        text: Select an Organizational Unit
      load:
        organizational_units: ous
      save: ou
    ```
    """
    return select(text, organizational_units, minimum=1, maximum=1)["first"]


@ldap_function
def exists(domain_controller: DomainController, **search_params: Union[str, int, None]) -> bool:
    """Search for organizational units using LDAP attributes & values to identify if any were found.

!!! question "Minimum Plugin Version: 5.2.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [OrganizationalUnit keys](../inputs_outputs#organizationalunit) & values to use in the search

??? success "Output"

    A boolean is outputted by this action

    - `true` if one or more organizational units are found

    - `false` if no organizational units are found


??? example

    Searching using a unique attribute:

    ```YAML linenums="1"
    - active_directory.organizational_units.exists:
        distinguishedName: "OU=Engineers,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

!!! tip "Wildcard Searching"

    It is possible to use `*` as a wildcard at the __end__ of search values, but this can make the search slow.

    """
    return bool(organizational_units.search(domain_controller, **search_params))
